/*
 * @file mesh_tracing.cpp
 * @brief Implementation of the paper "Real-Time Acoustics Simulation using
 *        Mesh-Tracing", Schiettecatte et al., 2003.
 * @author Kenji Inoue
 *
 * $Id: mesh_tracing.cpp 1841 2007-10-02 05:44:55Z SYSTEM $
 *
 */

#include <iostream>
#include <complex>
#include <vector>
#include <set>
#include <queue>
#include <stdlib.h>
#include <math.h>
#include <GL/glut.h>

using namespace std;

#define DELAUNAY_DEBUG 0
#define USE_GL         0
#define TARGET_SOURCE  8
#define LINEAR_INTERPOLATION 0



const double PI  = atan(1.0) * 4;
const double DT  = 10e-3; // 10msec
const double END = 50; // 50sec
const double F = 0.3; // frequency = 1Hz
const double IR_END = 5.0; // 5sec
//const double DT = 10e-9; // 10nsec
//const double END = 232.5e-6; // 232.5usec
//const double F = 500e3; // frequency = 500kHz

// constants for calculating impulse responses
const double THRESHOLD = 1e-6; // @todo: should be changed to how many times it reflected against the wall?
const double SOURCE_POWER = 1.0;
const double ATTENUATION_RATIO = 0.1; // loss factor per meter



struct Arg {
  int idx;
  double arg;

  bool operator<(const Arg &a) const {
    return (arg < a.arg);
  }
};

typedef complex<double> P;
struct T {
  int v1, v2, v3;
  P c;
};
typedef vector<T> V;



struct Power {
  double power;
  double time;

  bool operator<(const Power &a) const {
    return (time < a.time);
  }

  Power() {
    power = 0.0;
    time  = 0.0;
  }

  Power(double _power, double _time) {
    power = _power;
    time  = _time;
  }
};

struct Vertex {
  P p;
  set<int> to;
  vector< vector<Power> > powers; // e.g. vertex.powers[source_index].power
};

typedef vector<Vertex> Vertices;

struct Node {
  int v, from;
  double power;
  double length_so_far;

  Node(int _v, int _from, double _power, double _length_so_far) {
    v     = _v;
    from  = _from;
    power = _power;
    length_so_far = _length_so_far;
  }
};


Vertices vs;



#define EPS (1e-10)

double dot(P a, P b) {
  return (a.real() * b.real() + a.imag() * b.imag());
}

double cross(P a, P b) {
  return (a.real() * b.imag() - a.imag() * b.real());
}

P intersection_l(P a1, P a2, P b1, P b2) {
  P a = a2 - a1; P b = b2 - b1;
  return a1 + a * cross(b, b1-a1) / cross(b, a);
}

void compute_center_points(V &v, const P *p) {
  for (int i = 0; i < v.size(); i++) {
    if (v[i].v1 < 0) { continue; }

    P v1 = p[ v[i].v1 ];
    P v2 = p[ v[i].v2 ];
    P v3 = p[ v[i].v3 ];

//printf("<%d, %d, %d>\n", v[i].v1, v[i].v2, v[i].v3);

    P c1 = v1 + (v2 - v1) / 2.0;
    P c2 = v2 + (v3 - v2) / 2.0;
//      P c3 = v3 + (v1 - v3) / 2.0;
//printf("v1=[%lf,%lf], v2=[%lf,%lf], v3=[%lf,%lf]\n", v1.real(), v1.imag(), v2.real(), v2.imag(), v3.real(), v3.imag());

    P n1 = (v2 - c1) * P(0, 1);
    P n2 = (v3 - c2) * P(0, 1);
//      P n3 = (v1 - c3) * P(0, 1);

    v[i].c = intersection_l(c1, c1 + n1, c2, c2 + n2);
//printf("v[%d].c=[%lf,%lf]\n", i, v[i].c.real(), v[i].c.imag());
  }
}



void display(void) {
  glClear(GL_COLOR_BUFFER_BIT);
  glColor3f(0.0, 0.0, 0.0);

  for (int i = 0; i < vs.size(); i++) {
    P p1 = vs[i].p;

    for (set<int>::iterator it = vs[i].to.begin(); it != vs[i].to.end(); it++) {
      P p2 = vs[*it].p;
      glBegin(GL_LINES);
        glVertex3f( p1.real() * 90 + 5, p1.imag() * 90 + 5, 0.0 );
        glVertex3f( p2.real() * 90 + 5, p2.imag() * 90 + 5, 0.0 );
      glEnd();
    }
  }

  glColor3f(1.0, 0.0, 0.0);
  for (int i = 0; i < vs.size(); i++) {
    double x = vs[i].p.real() * 90 + 5;
    double y = vs[i].p.imag() * 90 + 5;

    glBegin(GL_POLYGON);
      glVertex2f(x - 0.5, y + 0.5);
      glVertex2f(x + 0.5, y + 0.5);
      glVertex2f(x + 0.5, y - 0.5);
      glVertex2f(x - 0.5, y - 0.5);
    glEnd();
  }

  glFlush();
  glutSwapBuffers();
}

void reshape(int width, int height) {
  glViewport(0, 0, (GLsizei)width, (GLsizei)height);
//  glMatrixMode(GL_PROJECTION);
//  glLoadIdentity();				/*単位行列代入*/
  glOrtho(0.0, 100, 0.0, 100, -1.0, 1.0);
//  glMatrixMode(GL_MODELVIEW);			/*モデル変換宣言*/
//  glLoadIdentity();				/*単位行列代入*/
}

void init_gl(int &argc, char** &argv) {
  /* initialize GLUT */
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB);
  glutInitWindowSize(800, 800);
  glutInitWindowPosition(100, 100);
  glutCreateWindow(argv[0]);

  /* assign the callback functions */
  glutIdleFunc(display);
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);

  glClearColor(1.0, 1.0, 1.0, 1.0); /* white color */
  glShadeModel(GL_FLAT);

  glutMainLoop();
}



bool write_vector_to(const char *filename, const vector<double> &v) {
  FILE *fp = fopen(filename, "w");
  if (!fp) { return false; }

  for (int i = 0; i < v.size(); i++) {
    fprintf(fp, "%.10lf\n", v[i]);
  }
  fclose(fp); fp = NULL;

  return true;
}

bool write_powers_to(const char *filename, const vector<Power> &powers) {
  FILE *fp = fopen(filename, "w");
  if (!fp) { return false; }

  for (int i = 0; i < powers.size(); i++) {
    fprintf(fp, "%.10lf %.10lf\n", powers[i].time, powers[i].power);
  }
  fclose(fp); fp = NULL;

  return true;
}

void generate_input_wave(vector<double> &in) {
  for (double t = 0; t < END; t += DT) {
    //double v = sin(t);
    double TPFT = 2 * PI * F * t; // 2 pi f t
    double v = sin(TPFT) * (0.5 - 0.5 * cos(TPFT / 5));
    in.push_back(v);
  }
}

void enqueue_vertices(queue<Node> &q, Vertices &vs, int source, const Node &n, const set<int> &bag) {
  // n   : current node
  // bag : vertex numbers to be added

  double max = 0.0;
  for (set<int>::iterator it = bag.begin(); it != bag.end(); it++) {
    int to = *it;
    double length = abs(vs[to].p - vs[n.v].p);
    if (length > max) { max = length; }
  }

  double total = 0.0;
  for (set<int>::iterator it = bag.begin(); it != bag.end(); it++) {
    int to = *it;
    double length = abs(vs[to].p - vs[n.v].p);
    total += max / length;
  }

  for (set<int>::iterator it = bag.begin(); it != bag.end(); it++) {
    int to = *it;

    double length = abs(vs[to].p - vs[n.v].p);
    double power = n.power * ((max / length) / total) * pow(1.0 - ATTENUATION_RATIO, length);
    double length_so_far = n.length_so_far + length;
    if (power > THRESHOLD) {
      q.push( Node(to, n.v, power, length_so_far) );
      vs[to].powers[source].push_back( Power(power, length_so_far) );
    } else {
      // too less power to push it
    }
  }
}

bool update_right(int *right_idx, Power *left, Power *right, const vector<Power> &powers) {
  (*right_idx)++;
  if (*right_idx < powers.size()) {
    *right = powers[*right_idx];
  } else {
    *right = Power(0.0, min(IR_END, END));
  }
  return true;
}

void render(vector<double> &out, const vector<double> &in, const Vertices &vs, int source, int listener) {
  if (in.size() < (int)(END / DT)) {
    printf("render(): error, in.size() is less than END / DT\n");
    return;
  }

  const int ir_end = (int)(min(IR_END, END) / DT);

//printf("\nimpulse response (from %d to %d)\n\n", source, listener);
  vector<double> ir(ir_end); // impulse response
  int right_idx = 0; // make it updated at first iteration
  // @todo: check the case powers[].size is less than 2
  Power left  = Power(0.0, 0.0),                // power would be 0.0 at time 0.0
        right = vs[listener].powers[source][0]; // make it updated at first iteration
  double dp; // delta power for interpolation
  bool updated_flag = false;
  for (int i = 0; i < ir_end; i++) {
    double t = i * DT;

    if (right.time < t) {
      left = right;
      updated_flag = update_right(&right_idx, &left, &right, vs[listener].powers[source]);
    }

    while ( abs(right.time - left.time) < DT ) {
      // combine the powers reached at the very same time
      left.power += right.power;
      updated_flag = update_right(&right_idx, &left, &right, vs[listener].powers[source]);
    }

    if (LINEAR_INTERPOLATION) {
      // linear interpolation
      if (updated_flag) {
        dp = (right.power - left.power) / (right.time - left.time);
      }

      // @todo: here the linear interpolation is employed
      //        maybe other, cosine interpolation for example, would be better
      ir[i] = left.power + dp * (t - left.time);

    } else {
      // no interpolation
      if (updated_flag) {
        ir[i] = left.power;
      } else {
        ir[i] = 0.0;
      }
    }

    if (updated_flag) { updated_flag = false; }
//if (i % 10 == 0) { printf("ir[%d/%lf] = %lf\n", i, i * DT, ir[i]); }
  }

  char filename[1024];
  sprintf(filename, "impulse_responses/%03d_to_%03d.ir-interpolated.data", source, listener);
  write_vector_to(filename, ir);

  const int end = (int)(END / DT);
  out.resize(end);
  for (int i = 0; i < end; i++) {
    out[i] = 0.0;

    int last = min(i, (int)(ir.size() - 1));
    for (int j = 0; j <= last; j++) {
      out[i] += in[i-j] * ir[j];
    }
  }
}

void write_energy_to(FILE *fp_energy, const vector<double> &v, const char *label) {
  double energy = 0.0; // energy
  for (int k = 0; k < v.size(); k++) {
    energy += pow(v[k], 2.0);
  }
  fprintf(fp_energy, "%s  %.10lf\n", label, energy);
}

void calculate_impulse_responses(Vertices &vs, int source) {
  // the output of this functions is powers data (impulse responses) only,
  // which is stored in vs[ALL].powers[source] vectors

  // values at the time 0.0
  for (int i = 0; i < vs.size(); i++) {
    if (source != i) {
      vs[i].powers[source].push_back( Power(0.0, 0.0) );
    } else {
      vs[i].powers[source].push_back( Power(SOURCE_POWER, 0.0) );
    }
  }

  queue<Node> q; // queue of <vertex, from> pairs
  q.push( Node(source, source, SOURCE_POWER, 0.0) );

  while ( !q.empty() ) {
    Node n = q.front();
    q.pop();

    if (n.v == n.from) {
      // V is the source vertex
      enqueue_vertices(q, vs, source, n, vs[n.v].to);

    } else {
      // V is not the source vertex

      P perp = (vs[n.v].p - vs[n.from].p) * P(0, 1); // perpendicular (normal) vector

      set<int> to_side, from_side;
      for (set<int>::iterator it = vs[n.v].to.begin(); it != vs[n.v].to.end(); it++) {
        int to = *it;

        if ( cross(perp, vs[to].p - vs[n.v].p) < 0 ) {
          to_side.insert(to);
        } else {
          from_side.insert(to);
        }
      }

      // @todo: is this condition okay? I bet it works as far as the room is rectangular, but...
      //        we may calculate it is the hull of the graph
      if (to_side.size() != 0) {
        // not wall
        enqueue_vertices(q, vs, source, n, to_side);

      } else {
        // wall
        enqueue_vertices(q, vs, source, n, from_side);
      }
    }
  }

  // sort by time
  for (int i = 0; i < vs.size(); i++) {
    sort(vs[i].powers[source].begin(), vs[i].powers[source].end());
  }

  // sum up the powers at the very same time
  for (int i = 0; i < vs.size(); i++) {
    vector<Power> pows( vs[i].powers[source].size() );

    int size = 0; // current size
    for (int j = 0; j < vs[i].powers[source].size(); j++) {
      pows[size] = vs[i].powers[source][j];

      if (size > 0 && abs(pows[size].time - pows[size-1].time) < EPS) {
        pows[size-1].power += pows[size].power;
      } else {
        size++;
      }
    }

    vs[i].powers[source].resize(size);
    for (int j = 0; j < size; j++) {
      vs[i].powers[source][j] = pows[j];
    }
  }
}



int main(int argc, char** argv) {
  int N = 10;

  if (argc > 1) { N = atoi(argv[1]); }
  srand(argc > 2 ? argv[2][0] : 1);



  printf("\n----- Generating the Input Wave -----\n\n");

  vector<double> in;
  generate_input_wave(in);
  write_vector_to("input.data", in);

  FILE *fp_energy = fopen("energy.txt", "w");
  write_energy_to(fp_energy, in, "input     ");



  P p[N];

  p[0].real() = 0.0; p[0].imag() = 0.0;
  p[1].real() = 0.0; p[1].imag() = 1.0;
  p[2].real() = 1.0; p[2].imag() = 0.0;
  p[3].real() = 1.0; p[3].imag() = 1.0;
  for (int i = 0; i < N-4; i++) {
    double x = (double)rand() / RAND_MAX;
    if (x == 0) { i--; continue; }

    p[i+4].real() = x;

    double y = (double)rand() / RAND_MAX;
    if (y == 0) { i--; continue; }
    p[i+4].imag() = y;
  }

printf("\n----- Vertices Randomly Distributed in (0,1) -----\n\n");
  for (int i = 0; i < N; i++) {
    printf("%3d : [%.10lf, %.10lf]\n", i, p[i].real(), p[i].imag());
  }

  // Delaunay Triangulation

  T tmp1 = {0, 1, 2}, tmp2 = {1, 2, 3};

  V v;
  v.push_back(tmp1);
  v.push_back(tmp2);

//  compute_center_points(v, p);

//  vector<int> tv;
  for (int i = 4; i < N; i++) {
    compute_center_points(v, p);

    set<int> bag;
    vector<int> tv;

    for (int j = 0; j < v.size(); j++) {
      if (v[j].v1 < 0) { continue; }

      P v1 = p[ v[j].v1 ];

      if ( abs(v[j].c - p[i]) < abs(v[j].c - v1) + EPS ) {
//printf("inserted.\n");
        tv.push_back(j);
        bag.insert( v[j].v1 );
        bag.insert( v[j].v2 );
        bag.insert( v[j].v3 );
      }
    }

    vector<Arg> sorted;

    P c = p[i];
    P anchor;
    bool anchor_flag = true;
    for (set<int>::iterator it = bag.begin(); it != bag.end(); it++) {
//printf("%d ", *it);

      if (anchor_flag) {
        anchor = p[*it];
        anchor_flag = false;
      }

      P a = (anchor - c) / abs(anchor - c);
      P b = (p[*it] - c) / abs(p[*it] - c);

      double arg = dot(a, b);
      double direction = cross(a, b);

      Arg tmp;
      tmp.idx = *it;
      if (direction > -EPS) {
        tmp.arg =  arg + 1.0;
      } else {
        tmp.arg = -arg - 1.0;
      }
      sorted.push_back(tmp);
    }
//printf("\n");
    sort(sorted.begin(), sorted.end());

    // delete
    for (int j = 0; j < tv.size(); j++) {
      v[ tv[j] ].v1 = -1;
    }

    // add
    for (int j = 0; j < sorted.size(); j++) {
//printf("[%d] idx=%d, arg=%lf\n", j, sorted[j].idx, sorted[j].arg);
      int a = j;
      int b = j + 1 != sorted.size() ? j + 1 : 0;

      T tmp = {sorted[a].idx, sorted[b].idx, i};
      v.push_back(tmp);
    }

/*
printf("Triangles:\n");
for (int i = 0; i < v.size(); i++) {
  if (v[i].v1 < 0) { continue; }

  int v1 = v[i].v1;
  int v2 = v[i].v2;
  int v3 = v[i].v3;

  printf("  [%3d] %3d-%3d-%3d (%lf,%lf)-(%lf,%lf)-(%lf,%lf)\n",
         i, v1, v2, v3, p[v1].real(), p[v1].imag(), p[v2].real(), p[v2].imag(), p[v3].real(), p[v3].imag());
}
*/
  }

if (DELAUNAY_DEBUG) { compute_center_points(v, p); }



  // path 2

  for (int i = 0; i < N; i++) {
    Vertex vx = { p[i] };
    vx.to.clear();

    vx.powers.clear();
    vx.powers.resize(N);
    for (int j = 0; j < N; j++) {
      vx.powers[j].clear();
    }

    vs.push_back(vx);
  }

  for (int i = 0; i < v.size(); i++) {
    if (v[i].v1 < 0) { continue; }

    int v1 = v[i].v1;
    int v2 = v[i].v2;
    int v3 = v[i].v3;

    // v1 <-> v2
    vs[v1].to.insert(v2);
    vs[v2].to.insert(v1);

    // v2 <-> v3
    vs[v2].to.insert(v3);
    vs[v3].to.insert(v2);

    // v3 <-> v1
    vs[v3].to.insert(v1);
    vs[v1].to.insert(v3);
  }

if (USE_GL) { init_gl(argc, argv); }



  // after here, the output only depends on vs

  printf("\n----- Generateing the Impulse Responses and Output Waves -----\n\n");

  vector<double> out; // buffer
  for (int source = 0; source < vs.size(); source++) {
    if (TARGET_SOURCE >= 0 && source != TARGET_SOURCE) { continue; }

    printf("Calculating impulse responses from %d ...\n", source);
    calculate_impulse_responses(vs, source);

    for (int listener = 0; listener < vs.size(); listener++) {
      printf(" processing from %d to %d ...\n", source, listener);

      // @todo: check if impulse_responses/out directories exist (otherwise it throws segv)

      char ir_filename[1024];
      sprintf(ir_filename, "impulse_responses/%03d_to_%03d.ir.data", source, listener);

      char out_filename[1024];
      sprintf(out_filename, "out/%03d_to_%03d.out.data", source, listener);

      char energy_label[1024];
      sprintf(energy_label, "%03d_to_%03d", source, listener);



      write_powers_to(ir_filename, vs[listener].powers[source]);

      // render the output from impulse response

      render(out, in, vs, source, listener);

      // write the out

      write_vector_to(out_filename, out);

      // write the total energy of out

      write_energy_to(fp_energy, out, energy_label);
    }
  }

  fclose(fp_energy);
  printf("\nWrote 'energy.txt'.\n\n");

  return 0;
}
